######################################################################
##
## Copyright (C) 2008,  Simon Kagstrom
##
## Author:        Simon Kagstrom <simon.kagstrom@gmail.com>
## Description:   Bookmarks in functions/files
##
## Licensed under the terms of GNU General Public License version 2 or
## later. See COPYING file distributed with Dissy for full text of the
## license.
##
######################################################################

class Bookmark:
    def __init__(self, comment, address):
        self.comment = comment
        self.location = location

class BookmarkSet:
    def __init__(self, filename):
        self.bookmarks = []
        self.md5sum = 0





######################################################################
##
## Copyright (C) 2006,  Blekinge Institute of Technology
##
## Author:        Simon Kagstrom <simon.kagstrom@gmail.com>
## Description:   Base-class for architecture handling
##
## Licensed under the terms of GNU General Public License version 2
## (or later, at your option). See COPYING file distributed with Dissy
## for full text of the license.
##
######################################################################
class Architecture:
    """
    Architecture base class. Inherit this to implement
    architecture-specific handling (see intel.py and mips.py)
    """
    def __init__(self, arch_jumps = [], arch_calls = [], arch_conditionflagsetters = [],
        arch_conditionflagusers = []):
        self.jumps = {}
        self.calls = {} # To handle reverse-engineering
        self.conditionflagsetters = {}
        self.conditionflagusers = {}
        for s in arch_jumps:
            self.jumps[s.strip()] = True
        for s in arch_calls:
            self.calls[s.strip()] = True
        for s in arch_conditionflagsetters:
            self.conditionflagsetters[s.strip()] = True
        for s in arch_conditionflagusers:
            self.conditionflagusers[s.strip()] = True

    def isJump(self, insn):
        "Returns true if this instruction is a jump"
        return insn in self.jumps

    def isCall(self, insn):
        "Returns true if this instruction is a call"
        return insn in self.calls

    def isConditionFlagSetter(self, insn):
        "Returns true if this instruction sets the condition flags"
        return insn in self.conditionflagsetters

    def isConditionFlagUser(self, insn):
        "Returns true if this instruction uses the condition flags"
        return insn in self.conditionflagusers

    def getJumpDestination(self, insn, args):
        """Parse the instruction to return the jump destination. The
        base class only tries to convert the argument to a number. See
        mips.py for a more advanced translation.
        """
        try:
            return long(args, 16)
        except ValueError:
            pass
        return None

from dissy import mips, intel, ppc, arm

def getArchitecture(archStr):
    if archStr == "intel":
        return intel.IntelArchitecture()
    elif archStr == "x86-64":
        return intel.IntelArchitecture()
    elif archStr == "i8086":
        return intel.IntelArchitecture()
    if archStr == "mips":
        return mips.MipsArchitecture()
    if archStr == "ppc":
        return ppc.PpcArchitecture()
    if archStr == "powerpc":
        return ppc.PpcArchitecture()
    if archStr == "arm":
        return arm.ArmArchitecture()
    if archStr == "arm26":
        return arm.ArmArchitecture()
    return Architecture([])



######################################################################
##
## Copyright (C) 2006,  Blekinge Institute of Technology
##
## Author:        Simon Kagstrom <simon.kagstrom@gmail.com>
## Description:   String entity
##
## Licensed under the terms of GNU General Public License version 2
## (or later, at your option). See COPYING file distributed with Dissy
## for full text of the license.
##
######################################################################
from dissy.Entity import Entity

class StrEntity(Entity):
    def __init__(self, fn, string):
        self.string = string
        self.function = fn

    def getFunction(self):
        return self.function

    def __str__(self):
        return self.string



######################################################################
##
## Copyright (C) 2006,  Blekinge Institute of Technology
##
## Author:        Simon Kagstrom <simon.kagstrom@gmail.com>
## Description:   Arm arch specific stuff
##
## Licensed under the terms of GNU General Public License version 2
## (or later, at your option). See COPYING file distributed with Dissy
## for full text of the license.
##
######################################################################
import sys, architecture
from dissy.architecture import Architecture

arm_jumps = ['b',
             'bcc',
             'bl',
             'ble',
             'bne',
             'bleq',
             'blt',
             'bgt',
             'beq',
             'bcs',
             ]
arm_calls = ['bl']
arm_conditionflag_setters = ['cmp'] + \
    [i + "s" for i in
    ['mul', 'mla', 'umull', 'umlal', 'smull',
    'smlal', 'mov', 'mvn', 'asr', 'lsl', 'lsr', 'ror', 'rrx',] #TODO more
    ]
arm_conditionflag_users = ['']

arm_instr_descriptions = {
    'cmp': 'Compare two values, and sets the condition flags',
    'adc': 'Add with carry',
    'add': 'Add',
    'sub': 'Subtract',
    'and': 'Logical and',
    'bic': 'Bit Clear',
    'bal': 'Unconditional Branch',
    'blal': 'Unconditional Branch and Link',
    'mov': 'Move',
    'mvn': 'Move and negate (XOR 0xFFFFFFFF)',
    'bx': """Branch and eXchange
PC := Rm""",
    'bl': """Branch with Link
LR := Address of next instruction, PC := label""",
    'push': """Push on the stack.
Canonical form of "stmdb SP!, <reglist>\"""",
    'pop': """Pop from the stack.
Canonical form of "ldm SP!, <reglist>\"""",
    'asr': 'Arithmetic Shift Right',
    'mul': 'Multiply',
    'muls': 'Multiply and set condition flags',
    'mla': 'Multiply and Accumulate',
    'mls': 'Multiply and Subtract',
    'smull': """Signed Multiply Long
%(arg1)s,%(arg2)s := signed(%(arg3)s * %(arg4)s)""",
    'ldr': 'Load (from memory to register)',
    'str': 'Store (from register to memory)',
    'lsl': 'Logical Shift Left',
    'lsls': 'Logical Shift Left and set condition flags',
    'lsr': 'Logical Shift Right',
    'lsr': 'Logical Shift Right and set condition flags',
    'rsb': 'Reverse Subtract',
}

arm_conditions = {
    'cs': 'Carry Set',
    'cc': 'Carry Clear',
    'eq': 'Equal (Zero Set)',
    'ne': 'Not Equal (Zero Clear)',
    'vs': 'Overflow Set',
    'vc': 'Overflow Clear',
    'gt': 'Signed Greater Than',
    'lt': 'Signed Less Than',
    'ge': 'Signed Greater than or Equal',
    'le': 'Signed Less than or Equal',
    'pl': 'Plus (Positive)',
    'mi': 'Minus (Negative)',
    'hi': 'Unsigned Higher Than',
    'lo': 'Unsigned Lower Than',
    'hs': 'Unsigned Higher or Same',
    'ls': 'Unsigned Lower or Same',
    }

arm_lists_inited = False
if not arm_lists_inited:
    conditional_instructions = {
        'b': """Branch on %s
PC := label, label is this instruction +/-32Mb""",
        'bx': """Branch and eXchange on %s
PC := Rm
Change to Thumb mode if Rm[0] is 1, change to ARM mode if Rm[0] is 0""",
        'bl': """Branch and Link on %s""",
        'mov': """Move on %s""",
        'add': """Add on %s""",
        'sub': """Subtract on %s""",
        }

    for i in conditional_instructions:
        for c in arm_conditions:
            arm_instr_descriptions[i + c] = conditional_instructions[i] % (arm_conditions[c])
            arm_conditionflag_users += [i + c]
    arm_lists_inited = True

class ArmArchitecture(architecture.Architecture):
    def __init__(self):
        architecture.Architecture.__init__(self, arm_jumps, arm_calls,
            arm_conditionflag_setters, arm_conditionflag_users)

    def getInstructionInfo(self, instruction):
        opcode = instruction.getOpcode()
        args = str(instruction.getArgs())

        args_list = args.split(", ")
        args_dict = dict(
            zip(['arg' + str(i) for i in range(1, len(args_list)+1)],
                args_list))
        description = arm_instr_descriptions.get(instruction.getOpcode(), '')

        return {'shortinfo': opcode + " " + args,
                'description': description % args_dict,
                }

    def parseArguments(self, instr):
        """Parses an argument string, into a two-tuple, containing:
            * A list of the registers read in this instruction
            * A list of the registers written to in this instruction"""

        def parseComSepList(lstr):
            ret = []
            tmp = ""
            stack = []
            for c in lstr:
                if c in "{[":
                    stack += [c]
                    tmp += c
                elif c in "]}":
                    tmp += c
                    c2 = stack.pop()
                    if c == "]" and c2 == "[" or \
                        c == "}" and c2 == "{":
                        ret += [tmp]
                        tmp = ""
                    else:
                        raise ValueError("unbalanced parenthesis")
                elif c == ' ':
                    pass
                elif stack == [] and c == ',':
                    ret += [tmp]
                    tmp = ""
                elif stack == []:
                    tmp += c
                else:
                    tmp += c
            if tmp:
                ret += [tmp]
            return ret
        def isRegister(s):
            if s in ['r' + str(i) for i in range(0, 16)]: #r0..r15
                return True
            if s in ['sl', #r10 = gcc "got pointer"?
                    'fp', #r11 = gcc frame pointer
                    'ip', #r12 = gcc "scratch register"?
                    'sp', #r13, stack pointer
                    'lr', #r14, link register
                    'pc']: #r15, program counter
                return True
            return False
        def isValue(s):
            if s[:1] == '#':
	        if s[1:].isdigit():
                    return True
            return False
        regwrite = []
        regread = []
        values = []
        args = parseComSepList(instr.args)
        values = [int(a[1:]) for a in args if isValue(a)]

        if instr.getOpcode()[:3] == 'cmp':            
            regread = [a for a in args if isRegister(a)]
        elif instr.getOpcode()[:3] in ['sub', 'add', 'lsl', 'asr', 'rsb', 'mov', 'and', 'mvn']:
            regwrite = [args[0]]
            regread = [a for a in args[1:] if isRegister(a)]
        #branches
        elif instr.getOpcode() in ['b' + c for c in arm_conditions.keys() + ['']]:
            regwrite = ['pc']
        elif instr.getOpcode() in ['bl' + c for c in arm_conditions.keys() + ['']]:
            regwrite = ['pc', 'lr']
        elif instr.getOpcode() in ['bx' + c for c in arm_conditions.keys() + ['']]:
            regwrite = ['pc']
            regread = [args[0]]
        elif instr.getOpcode() in ['blx' + c for c in arm_conditions.keys() + ['']]:
            regwrite = ['pc']
            regread = isRegister(args[0]) and [args[0]] or []
        #load
        elif instr.getOpcode() in ['ldr' + c for c in arm_conditions.keys() + ['']]:
            regwrite = [args[0]]
            if args[1].startswith('['):
                offsetl = parseComSepList(args[1][1:-1])
                regread = [r for r in offsetl if isRegister(r)]
        #store
        elif instr.getOpcode() in ['str' + c for c in arm_conditions.keys() + ['']]:
            regread = [args[0]]
            if args[1].startswith('['):
                offsetl = parseComSepList(args[1][1:-1])
                regread += [r for r in offsetl if isRegister(r)]
        #push
        elif instr.getOpcode() in ['push' + c for c in arm_conditions.keys() + ['']]:
            regwrite = ['sp']
            regread = ['sp']
            reglist = parseComSepList(args[0][1:-1])
            regread += reglist
        elif instr.getOpcode() in ['pop' + c for c in arm_conditions.keys() + ['']]:
            regwrite = ['sp']
            regread = ['sp']
            reglist = parseComSepList(args[0][1:-1])
            regwrite += reglist
        elif instr.getOpcode() in ['smull' + c for c in arm_conditions.keys() + ['']]:
            regwrite = [args[0], args[1]]
            regread = [args[2], args[3]]
        elif instr.getOpcode() in ['mla' + c for c in arm_conditions.keys() + ['']]:
            regwrite = [args[0]]
            regread = args[1:]
        elif instr.getOpcode() in ['mul' + c for c in arm_conditions.keys() + ['']] + \
        ['muls' + c for c in arm_conditions.keys() + ['']]:
            regwrite = [args[0]]
            regread = [args[1], args[2]]
        elif instr.getOpcode() == '.word':
            return ([], [], [])
        else:
            raise ValueError("Unknown instruction opcode: " + str(instr))

        return (regread, regwrite, values)



######################################################################
##
## Copyright (C) 2006,  Blekinge Institute of Technology
##
## Author:        Simon Kagstrom <simon.kagstrom@gmail.com>
## Description:   A function
##
## Licensed under the terms of GNU General Public License version 2
## (or later, at your option). See COPYING file distributed with Dissy
## for full text of the license.
##
######################################################################
import re, os, cgi

from dissy.Config import config
from dissy.Entity import Entity, AddressableEntity
from dissy.Instruction import Instruction
from dissy.JumpStreamHandler import *
from dissy.StrEntity import StrEntity

ADDRESS_REGEXP  = "[0-9,a-f,A-F]+"
ENCODING_REGEXP = ADDRESS_REGEXP + "[ ]"
INSN_REGEXP     = "[0-9,a-z,A-Z,_,\-,\.,\+]+"
INSN_ARGS_REGEXP= "\**[a-z,A-Z,0-9,_,\,,\(,\),\%,\$,\[,\],!,#,\-, ,&,{,},\*,\+]+"

insnRegExp = re.compile("[ ]*(" + ADDRESS_REGEXP + "):[ \t]+((?:" + ENCODING_REGEXP +")*)[ \t]+(" + INSN_REGEXP + ")+[ \t]*(" + INSN_ARGS_REGEXP + ")*")

class Function(AddressableEntity):
    def __init__(self, fileContainer, address, label, size=0):
        AddressableEntity.__init__(self, address = address, endAddress = address + size, baseAddress = fileContainer.baseAddress)
        self.label = label
        self.all = []
        self.insns = []
        self.file = fileContainer

    def addInstruction(self, insn):
        self.insns.append(insn)
        self.all.append(insn)
        if insn.getAddress() > self.endAddress:
            self.setSize(insn.getAddress() - self.address)

    def getFile(self):
        return self.file

    def addOther(self, other):
        self.all.append(StrEntity(self, other))

    def lookup(self, address):
        last = self.insns[0]
        for insn in self.insns[1:]:
            if address >= last.getAddress() and address < insn.getAddress():
                return last
            last = insn
        return None

    def parse(self, try64bitWorkaround=False):
        """Parse the function."""
        count = 0
        start, end = self.getExtents()
        if try64bitWorkaround:
            if start & (1<<31):
                start = long(start) | 0xffffffff00000000
            if end & (1<<31):
                end = long(end) | 0xffffffff00000000
        s = "%s --wide --demangle --source --start-address=0x%Lx --stop-address=0x%Lx %s" % (config.objdump, start, end, self.file.filename)
        self.stream = os.popen(s)
        if self.stream == None:
            self.prepareParseFunction()
        self.instructions = []
        self.all = []
        firstNonEmpty=False
        for line in self.stream:
            # Weed away some unneeded stuff
            if line.startswith("Disassembly of section ") or line.startswith("%s: " % (self.file.filename)):
                continue
            if not firstNonEmpty and line.strip() == "":
                continue

            firstNonEmpty=True
            r = insnRegExp.match(line)
            if r != None:
                insn = Instruction(self, long("0x" + r.group(1),16), r.group(2), r.group(3), r.group(4))
                self.addInstruction(insn)
                count = count + 1
            else:
                self.addOther(cgi.escape(line))
        self.stream.close()

        if count == 0 and try64bitWorkaround == False:
            # If we couldn't add anything interesting, try the 64-bit
            # workaround (for e.g., MIPS). See http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=369343
            return self.parse(try64bitWorkaround = True)

    def link(self):
        """
        Link the local jumps in this function. External jumps are
        linked to the function they reside in.
        """
        unresolved = []
        for insn in self.insns:
            if insn.hasLink() and not insn.link():
                unresolved.append(insn)

        positive, negative = self.getJumpDicts()

        # Fill in the jumps, forwards...
        self.fillInJumpStreams(positive, False)
        # ... and backwards
        self.all.reverse()
        self.fillInJumpStreams(negative, True)
        self.all.reverse()

        return unresolved

    def getInstructions(self):
        return self.insns

    def getAll(self):
        return self.all

    def getJumpDicts(self):
        """
        Get jump dictionaries for the forward and backward jumps.
        """
        positive = {}
        negative = {}
        for insn in self.insns:
            other = insn.getOutLink()
            if isinstance(other, Instruction) and other != insn:
                if other.getAddress() < insn.getAddress():
                    # Jump from insn to other BACKWARDS
                    negative[insn.getAddress()] = (insn, other)
                else:
                    # Jump from insn to other FORWARDS
                    positive[insn.getAddress()] = (insn, other)
        return positive, negative

    def fillInJumpStreams(self, jumpDict, left):
        """
        Fill in the jump streams for a dictionary of start
        addresses. Specify if the left or right streams should be
        generated.
        """
        jumpStreamHandler = JumpStreamHandler()
        for insn in self.all:
            if isinstance(insn, Instruction):
                # Something starts on this address
                if jumpDict.has_key(insn.getAddress()):
                    stream = jumpStreamHandler.alloc()
                    if stream != None:
                        stream.start(jumpDict[insn.getAddress()] )
            jumpStreamHandler.update(insn)
            if left == True:
                insn.left_state = jumpStreamHandler.getStateTuple()
            else:
                insn.right_state = jumpStreamHandler.getStateTuple()


    def __str__(self):
        out = str("0x%08x" % self.address) + " " + str(self.label) + ":\n"
        for insn in self.insns:
            out = out + str(insn) + "\n"
        return out



######################################################################
##
## Copyright (C) 2006,  Blekinge Institute of Technology
##
## Author:        Simon Kagstrom <simon.kagstrom@gmail.com>
## Description:   Describes data symbols
##
## Licensed under the terms of GNU General Public License version 2
## (or later, at your option). See COPYING file distributed with Dissy
## for full text of the license.
##
######################################################################
import re, os, cgi, curses.ascii

from dissy.Config import config
from dissy.Entity import Entity, AddressableEntity
from dissy.StrEntity import StrEntity

class DataBase(AddressableEntity):
    def __init__(self, fileContainer, address, label, size=0):
        AddressableEntity.__init__(self, address = address, endAddress = address + size, baseAddress = fileContainer.baseAddress)
        self.label = label
        self.file = fileContainer
        self.data = []
        self.string = ""
        self.stream = None
        self.type = "Unknown"

    def parse(self):
        pass

    def toNumericValue(self):
        val = 0
        if len(self.data) in (1,2,4,8):
            for i in range(0, len(self.data)):
                val = val + self.data[i] << (len(self.data) - i)*8
        return val

    def toBytes(self):
        return self.data

    def toString(self):
        return self.string

class Data(DataBase):
    def __init__(self, fileContainer, address, label, size=0):
        DataBase.__init__(self, fileContainer, address, label, size)
        self.type = "Initialized data"

    def parse(self):
        extents = self.getExtents()
        s = "%s --wide --demangle --full-contents --start-address=0x%x --stop-address=0x%x %s" % (config.objdump, extents[0], extents[1], self.file.filename)
        self.stream = os.popen(s)
        for line in self.stream:
            # Weed away some unneeded stuff
            if line.startswith("Contents of section ") or line.startswith("%s: " % (self.file.filename)):
                continue
            if line.strip() == "":
                continue

            words = line.split()
            for word in words[1:max(4, len(words))]:
                for i in range(0, len(word), 2):
                    try:
                        val = int(word[i:i+2], 16)
                        self.data.append(val)
                    except:
                        # For "short" data
                        continue
                    if curses.ascii.isprint(val):
                        self.string += "%c" % curses.ascii.ascii(val)
                    else:
                        self.string += "."
        self.stream.close()

class Bss(DataBase):
    def __init__(self, fileContainer, address, label, size=0):
        DataBase.__init__(self, fileContainer, address, label, size)
        self.type = "Uninitialized data"

    def parse(self):
        size = self.getSize()
        self.data = [ 0 ] * size
        self.string = "." * size

class RoData(Data):
    def __init__(self, fileContainer, address, label, size=0):
        Data.__init__(self, fileContainer, address, label, size)
        self.type = "Read-only data"
